//
//  RootViewController.m
//  liveshare
//
//  Created by Michael Heyeck on 10/20/10.
//  Copyright 2010 Fair Oaks Labs, Inc. All rights reserved.
//

#import <fcntl.h>
#import <unistd.h>
#import "liveshareAppDelegate.h"
#import "RootViewController.h"


@interface RootViewController ()

@property (nonatomic, retain) NSArray* fns;

- (void)syncOrderedList:(NSArray*)a withTarget:(NSArray*)b returnInserts:(NSMutableArray*)i andDeletes:(NSMutableArray*)d;
- (void)updateFns;
- (void)startMonitor;
- (void)stopMonitor;
- (void)restartMonitor;

@end


@implementation RootViewController

@synthesize fns;


#pragma mark -
#pragma mark View lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
	
	self.title = @"Documents";
	
	self.navigationItem.leftBarButtonItem = [[[UIBarButtonItem alloc] initWithTitle:@"Start" style:UIBarButtonItemStyleBordered target:self action:@selector(restartMonitor)] autorelease];
	self.navigationItem.rightBarButtonItem = [[[UIBarButtonItem alloc] initWithTitle:@"Stop" style:UIBarButtonItemStyleBordered target:self action:@selector(stopMonitor)] autorelease];

	liveshareAppDelegate* delegate = (liveshareAppDelegate*) [[UIApplication sharedApplication] delegate];
	self.fns = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:[delegate applicationDocumentsDirectory]
																   error:NULL];
	
	[self startMonitor];
}


/*
- (void)viewWillAppear:(BOOL)animated {
    [super viewWillAppear:animated];
}
*/
/*
- (void)viewDidAppear:(BOOL)animated {
    [super viewDidAppear:animated];
}
*/
/*
- (void)viewWillDisappear:(BOOL)animated {
	[super viewWillDisappear:animated];
}
*/
/*
- (void)viewDidDisappear:(BOOL)animated {
	[super viewDidDisappear:animated];
}
*/

/*
 // Override to allow orientations other than the default portrait orientation.
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation {
	// Return YES for supported orientations.
	return (interfaceOrientation == UIInterfaceOrientationPortrait);
}
 */

#pragma mark -
#pragma mark Memory management

- (void)didReceiveMemoryWarning
{
    // Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
    
    // Relinquish ownership any cached data, images, etc that aren't in use.
}


- (void)viewDidUnload
{
    // Relinquish ownership of anything that can be recreated in viewDidLoad or on demand.
    // For example: self.myOutlet = nil;
}


- (void)dealloc
{
	[self stopMonitor];
	[fns release];
    [super dealloc];
}

#pragma mark -
#pragma mark Table view data source

// Customize the number of sections in the table view.
- (NSInteger)numberOfSectionsInTableView:(UITableView*)tableView
{
    return 1;
}


// Customize the number of rows in the table view.
- (NSInteger)tableView:(UITableView*)tableView numberOfRowsInSection:(NSInteger)section
{
    return [self.fns count];
}


// Customize the appearance of table view cells.
- (UITableViewCell*)tableView:(UITableView*)tableView cellForRowAtIndexPath:(NSIndexPath*)indexPath
{
    static NSString* CellIdentifier = @"Cell";
    
    UITableViewCell* cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    if (cell == nil)
	{
        cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifier] autorelease];
    }
    
    // Configure the cell...
    cell.textLabel.text = [self.fns objectAtIndex:indexPath.row];
	
    return cell;
}


// Override to support conditional editing of the table view.
- (BOOL)tableView:(UITableView*)tableView canEditRowAtIndexPath:(NSIndexPath*)indexPath
{
    return YES;
}


// Override to support editing the table view.
- (void)tableView:(UITableView*)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath*)indexPath
{
    
    if (editingStyle == UITableViewCellEditingStyleDelete)
	{
		// Remove the file
		NSString* base = [(liveshareAppDelegate*)[[UIApplication sharedApplication] delegate] applicationDocumentsDirectory];
		NSString* path = [base stringByAppendingPathComponent:[self.fns objectAtIndex:indexPath.row]];
		[[NSFileManager defaultManager] removeItemAtPath:path error:NULL];
    }
}


/*
// Override to support rearranging the table view.
- (void)tableView:(UITableView *)tableView moveRowAtIndexPath:(NSIndexPath *)fromIndexPath toIndexPath:(NSIndexPath *)toIndexPath {
}
*/


/*
// Override to support conditional rearranging of the table view.
- (BOOL)tableView:(UITableView *)tableView canMoveRowAtIndexPath:(NSIndexPath *)indexPath {
    // Return NO if you do not want the item to be re-orderable.
    return YES;
}
*/

#pragma mark -
#pragma mark Table view delegate

- (void)tableView:(UITableView*)tableView didSelectRowAtIndexPath:(NSIndexPath*)indexPath
{
	[self.tableView deselectRowAtIndexPath:indexPath animated:YES];
}

#pragma mark -
#pragma mark Extension methods

- (void)syncOrderedList:(NSArray*)a withTarget:(NSArray*)b returnInserts:(NSMutableArray*)i andDeletes:(NSMutableArray*)d
{
	NSInteger x=0, y=0, la=[a count], lb=[b count]; [i removeAllObjects]; [d removeAllObjects];
	while ((x < la) || (y < lb))
	{
		if (y >= lb)
		{
			[d addObject:[NSNumber numberWithInteger:x]]; x++;
		}
		else if (x >= la)
		{
			[i addObject:[NSArray arrayWithObjects:[NSNumber numberWithInteger:y],[b objectAtIndex:y],nil]]; y++;
		}
		else
		{
			NSComparisonResult r = [[a objectAtIndex:x] localizedCompare:[b objectAtIndex:y]];
			if (r == NSOrderedAscending)
			{
				[d addObject:[NSNumber numberWithInteger:x]]; x++;
			}
			else if (r == NSOrderedDescending)
			{
				[i addObject:[NSArray arrayWithObjects:[NSNumber numberWithInteger:y],[b objectAtIndex:y],nil]]; y++;
			}
			else
			{
				x++; y++;
			}
		}
	}
}


- (void)updateFns
{
	// Fetch new directory contents
	liveshareAppDelegate* delegate = (liveshareAppDelegate*) [[UIApplication sharedApplication] delegate];
	NSArray* target = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:[delegate applicationDocumentsDirectory]
																		  error:NULL];

	// Calculate necessary synchronization operations
	NSMutableArray* i = [NSMutableArray arrayWithCapacity:5]; NSMutableArray* d = [NSMutableArray arrayWithCapacity:5];
	[self syncOrderedList:self.fns withTarget:target returnInserts:i andDeletes:d];
	
	// Swap FN list
	self.fns = target;
	
	// Synchronize table (Sure, we could just reload, but why not be pretty about it?)
	for (NSInteger idx = [i count]-1; idx >= 0; idx--)
	{
		NSIndexPath* path = [NSIndexPath indexPathForRow:[[[i objectAtIndex:idx] objectAtIndex:0] integerValue] inSection:0];
		[i replaceObjectAtIndex:idx withObject:path];
	}
	for (NSInteger idx = [d count]-1; idx >= 0; idx--)
	{
		NSIndexPath* path = [NSIndexPath indexPathForRow:[[d objectAtIndex:idx] integerValue] inSection:0];
		[d replaceObjectAtIndex:idx withObject:path];
	}
	[self.tableView beginUpdates];
	[self.tableView insertRowsAtIndexPaths:i withRowAnimation:UITableViewRowAnimationFade];
	[self.tableView deleteRowsAtIndexPaths:d withRowAnimation:UITableViewRowAnimationFade];
	[self.tableView endUpdates];
}


- (void)startMonitor
{
	// One ping only
    if (_src != NULL) return;

	// Fetch pathname of the directory to monitor
	liveshareAppDelegate* delegate = (liveshareAppDelegate*) [[UIApplication sharedApplication] delegate];
	NSString* docPath = [delegate applicationDocumentsDirectory];
	if (!docPath) return;
	
	// Open an event-only file descriptor associated with the directory
    int dirFD = open([docPath fileSystemRepresentation], O_EVTONLY);
	if (dirFD < 0) return;
	
	// Get the main thread's serial dispatch queue (since we'll be updating the UI)
	dispatch_queue_t queue = dispatch_get_main_queue();
	if (!queue)
	{
		close(dirFD);
		return;
	}
	
	// Create a dispatch source to monitor the directory for writes
	_src = dispatch_source_create(DISPATCH_SOURCE_TYPE_VNODE,	// Watch for certain events on the VNODE spec'd by the second (handle) argument
								  dirFD,						// The handle to watch (the directory FD)
								  DISPATCH_VNODE_WRITE,			// The events to watch for on the VNODE spec'd by handle (writes)
								  queue);						// The queue to which the handler block will ultimately be dispatched
	if (!_src)
	{
		close(dirFD);
		return;
	}
	
	// Set the block to be submitted in response to an event
	__block id unretained_self = self;
	dispatch_source_set_event_handler(_src, ^{[unretained_self updateFns];});

	// Set the block to be submitted in response to source cancellation
	dispatch_source_set_cancel_handler(_src, ^{close(dirFD);});
	
	// Unsuspend the source s.t. it will begin submitting blocks
	dispatch_resume(_src);
}


- (void)stopMonitor
{
	if (_src)
	{
		// Stop the source from submitting further blocks (and close the underlying FD)
		dispatch_source_cancel(_src);
		
		// Release the source
		dispatch_release(_src);
		_src = NULL;
	}
}


- (void)restartMonitor
{
	[self stopMonitor];
	[self startMonitor];
	[self updateFns];
}

@end
